#include "process.h"

#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include "models.h"

int blist_app_user_get(const blist_app *ctx) {
  blist_user *user = blist_user_get_by_local(ctx->handle, getuid());

  if (user) {
    printf("User: %s\nID: %lu\n", (char *)user->name, user->id);
    size_t lists_len = 0;
    blist_list **lists =
        blist_list_get_all_by_user(ctx->handle, user->id, &lists_len);
    printf("Lists: %lu\n", lists_len);
    int i = 0;
    for (; i < lists_len; i++)
      blist_list_deinit(lists[i]);
  } else
    blist_print(ctx->log_ctx, BLIST_LOG_ERROR, "No user for '%s' found.\n",
                getlogin());
  return 0;
}

int blist_app_user_create(const blist_app *ctx) {
  blist_print(ctx->log_ctx, BLIST_LOG_INFO,
              (blist_assure_user(ctx->handle, getuid(), getlogin()))
                  ? "User created.\n"
                  : "User already exists.\n");
  return 0;
}

int blist_app_user_delete(const blist_app *ctx) {
  blist_user *user = blist_user_get_by_local(ctx->handle, getuid());
  if (user) {
    blist_user_delete(ctx->handle, user);
    printf("User deleted.\n");
  } else
    blist_print(ctx->log_ctx, BLIST_LOG_ERROR,
                "User not found, no changes done.\n");
  return 0;
}

int blist_app_list_add(const blist_app *ctx) {
  blist_list_context *list_ctx = ctx->cmd_ctx;
  if (!list_ctx->stages_len) {
    blist_print(ctx->log_ctx, BLIST_LOG_ERROR, "Missing stages.\n");
    return 1;
  }
  blist_assure_user(ctx->handle, getuid(), getlogin());
  blist_user *user = blist_user_get_by_local(ctx->handle, getuid());
  blist_list *list =
      blist_list_get_by_user_and_name(ctx->handle, user->id, list_ctx->name);
  if (!list)
    list = blist_list_init();
  list->name = (unsigned char *)strdup(list_ctx->name);
  if (list->desc)
    list->desc = (unsigned char *)strdup(list_ctx->desc);
  list->is_preset = false;
  blist_list_save(ctx->handle, list, user);
  printf("Added list '%s'\n", (char *)list->name);
  int i = 0;
  for (; i < list_ctx->stages_len; i++) {
    blist_stage *stage = blist_stage_init();
    stage->name = (unsigned char *)strdup(list_ctx->stages[i]);
    stage->list_id = list->id;
    blist_stage_save(ctx->handle, stage);
    blist_stage_deinit(stage);
  }
  return 0;
}

int blist_app_list_edit(const blist_app *ctx) {
  blist_print(ctx->log_ctx, BLIST_LOG_ERROR, "Not implemented\n");
  return 0;
}

int blist_app_list_get(const blist_app *ctx) {
  blist_list_context *list_ctx = ctx->cmd_ctx;
  blist_user *user = blist_user_get_by_local(ctx->handle, getuid());

  size_t lists_len = 0;
  blist_list **lists = NULL;
  if (list_ctx->name) {
    blist_list *list =
        blist_list_get_by_user_and_name(ctx->handle, user->id, list_ctx->name);
    if (list) {
      lists_len = 1;
      lists = malloc(lists_len * sizeof(blist_list));
      memcpy(lists[0], list, sizeof(blist_list));
      blist_list_deinit(list);
    };
  } else
    lists = blist_list_get_all_by_user(ctx->handle, user->id, &lists_len);

  if (lists_len) {
    int i = 0;
    for (; i < lists_len; i++) {
      blist_list *list = lists[i];
      size_t tasks_len = 0;
      blist_task **tasks =
          blist_task_get_for_list(ctx->handle, list->id, &tasks_len);
      printf("Name: %s\n", (char *)list->name);
      printf("Description: %s\n", list->desc ? (char *)list->desc : "N/A");
      printf("Tasks: %ld\n", tasks_len);
      int j = 0;
      for (; j < tasks_len; j++)
        blist_task_deinit(tasks[j]);
      free(tasks);
      blist_list_deinit(list);
    }
    free(lists);
  } else if (list_ctx->name)
    blist_print(ctx->log_ctx, BLIST_LOG_ERROR, "List '%s' not found.\n",
                list_ctx->name);
  else
    blist_print(ctx->log_ctx, BLIST_LOG_WARNING, "No lists found.\n");
  return 0;
}

int blist_app_list_delete(const blist_app *ctx) {
  blist_list_context *list_ctx = ctx->cmd_ctx;
  blist_assure_user(ctx->handle, getuid(), getlogin());
  blist_user *user = blist_user_get_by_local(ctx->handle, getuid());
  if (!user) {
    blist_print(ctx->log_ctx, BLIST_LOG_ERROR, "User doesn't exist.\n");
    return 1;
  }

  blist_list *list =
      blist_list_get_by_user_and_name(ctx->handle, user->id, list_ctx->name);
  if (list) {
    size_t len = 0;
    blist_task **tasks = blist_task_get_for_list(ctx->handle, list->id, &len);
    int i = 0;
    for (; i < len; i++) {
      blist_task_delete(ctx->handle, tasks[i]);
      blist_task_deinit(tasks[i]);
    }
    blist_stage **stages =
        blist_stage_get_all_for_list(ctx->handle, list->id, &len);
    for (i = 0; i < len; i++) {
      blist_stage_delete(ctx->handle, stages[i]);
      blist_stage_deinit(stages[i]);
    }
    blist_list_delete(ctx->handle, list);
    blist_list_deinit(list);
    free(stages);
    free(tasks);
  } else {
    blist_print(ctx->log_ctx, BLIST_LOG_ERROR, "List '%s' not found.\n",
                list_ctx->name);
  }
  return 0;
}

int blist_app_task_add(const blist_app *ctx) {
  return blist_app_task_edit(ctx);
}

int blist_app_task_edit(const blist_app *ctx) {
  blist_task_context *task_ctx = ctx->cmd_ctx;
  blist_assure_user(ctx->handle, getuid(), getlogin());
  blist_user *user = blist_user_get_by_local(ctx->handle, getuid());
  blist_list *list = NULL;

  if (!user) {
    blist_print(ctx->log_ctx, BLIST_LOG_ERROR, "User doesn't exist.\n");
    return 1;
  }

  list = blist_list_get_by_user_and_name(ctx->handle, user->id, task_ctx->list);
  if (list) {
    size_t stages_len = 0;
    int i = 0;
    blist_stage **stages =
        blist_stage_get_all_for_list(ctx->handle, list->id, &stages_len);
    if (!task_ctx->stage) {
      blist_print(ctx->log_ctx, BLIST_LOG_ERROR,
                  "Stage missing/wrong, please pass one of: ");
      for (i = 0; i < stages_len; i++)
        fprintf(ctx->log_ctx->log_target, "%s ", (char *)stages[i]->name);
      fprintf(ctx->log_ctx->log_target, "\n");
    }
    long stage_id = -1;
    for (i = 0; i < stages_len; i++)
      if (strcmp((char *)stages[i]->name, task_ctx->stage) == 0)
        stage_id = stages[i]->id;

    if (stage_id == -1) {
      blist_print(ctx->log_ctx, BLIST_LOG_ERROR,
                  "Stage %s not found. Use one of: ", task_ctx->stage);
      for (i = 0; i < stages_len; i++)
        fprintf(ctx->log_ctx->log_target, "%s ", (char *)stages[i]->name);
      fprintf(ctx->log_ctx->log_target, "\n");
    }
    for (i = 0; i < stages_len; i++)
      blist_stage_deinit(stages[i]);
    free(stages);

    blist_task *task =
        blist_task_get_for_list_by_name(ctx->handle, list->id, task_ctx->name);
    if (!task)
      task = blist_task_init();
    if (task->name)
      free(task->name);
    task->name = (unsigned char *)strdup(task_ctx->name);
    if (task->desc)
      free(task->desc);
    if (task_ctx->desc)
      task->desc = (unsigned char *)strdup(task_ctx->desc);
    task->stage_id = stage_id;
    blist_task_save(ctx->handle, task);
    blist_task_deinit(task);
  } else
    blist_print(ctx->log_ctx, BLIST_LOG_ERROR, "List not found.\n");
  blist_user_deinit(user);
  if (list)
    blist_list_deinit(list);
  return 0;
}

int blist_app_task_get(const blist_app *ctx) {
  blist_task_context *task_ctx = ctx->cmd_ctx;
  blist_assure_user(ctx->handle, getuid(), getlogin());
  blist_user *user = blist_user_get_by_local(ctx->handle, getuid());
  if (!user) {
    blist_print(ctx->log_ctx, BLIST_LOG_ERROR, "User doesn't exist.\n");
    return 1;
  }
  blist_list *list =
      blist_list_get_by_user_and_name(ctx->handle, user->id, task_ctx->list);
  if (list) {
    if (task_ctx->name) {
      blist_task *task = blist_task_get_for_list_by_name(ctx->handle, list->id,
                                                         task_ctx->name);
      if (task) {
        printf("Name: %s\n", (char *)task->name);
        printf("Description: %s\n", task->desc ? (char *)task->desc : "N/A");
        blist_stage *stage = blist_stage_get_by_id(ctx->handle, task->stage_id);
        printf("Stage: %s\n", (char *)stage->name);
        blist_stage_deinit(stage);
        blist_task_deinit(task);
      } else
        blist_print(ctx->log_ctx, BLIST_LOG_ERROR, "Task not found.\n");
    } else {
      size_t tasks_len = 0;
      blist_task **tasks =
          blist_task_get_for_list(ctx->handle, list->id, &tasks_len);
      blist_print_list(ctx->handle, list, tasks, tasks_len);

      int i = 0;
      for (; i < tasks_len; i++)
        blist_task_deinit(tasks[i]);
      free(tasks);
    }
  } else
    blist_print(ctx->log_ctx, BLIST_LOG_ERROR, "list not found\n");
  blist_user_deinit(user);
  if (list)
    blist_list_deinit(list);
  return 0;
}

int blist_app_task_delete(const blist_app *ctx) {
  blist_task_context *task_ctx = ctx->cmd_ctx;
  blist_assure_user(ctx->handle, getuid(), getlogin());
  blist_user *user = blist_user_get_by_local(ctx->handle, getuid());
  if (!user) {
    blist_print(ctx->log_ctx, BLIST_LOG_ERROR, "User doesn't exist.\n");
    return 1;
  }
  blist_list *list =
      blist_list_get_by_user_and_name(ctx->handle, user->id, task_ctx->list);
  if (list) {
    if (task_ctx->name) {
      blist_task *task = blist_task_get_for_list_by_name(ctx->handle, list->id,
                                                         task_ctx->name);
      if (task) {
        blist_print(ctx->log_ctx, BLIST_LOG_WARNING, "Deleting task '%s'.\n",
                    (char *)task->name);
        blist_task_delete(ctx->handle, task);
        blist_task_deinit(task);
      } else
        blist_print(ctx->log_ctx, BLIST_LOG_ERROR, "Task not found.\n");
    } else
      blist_print(ctx->log_ctx, BLIST_LOG_ERROR,
                  "Task not found, only deletion by name is "
                  "currently supoprted.\n");
  } else
    blist_print(ctx->log_ctx, BLIST_LOG_ERROR, "Not implemented\n");
  return 0;
}