
#include <sqlite3.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/ioctl.h>
#include <unistd.h>

#include "blist.h"
#include "sql.h"

blist_user *blist_user_init() {
  blist_user *user = malloc(sizeof(blist_user));
  user->id = -1;
  user->name = NULL;
  user->local_id = -1;
  return user;
}

blist_user *blist_user_init_from_sql(sqlite3_stmt *stmt) {
  blist_user *user = blist_user_init();
  user->id = sqlite3_column_int(stmt, 0);
  if (sqlite3_column_bytes(stmt, 1) > 0)
    user->name = (unsigned char *)strdup((char *)sqlite3_column_text(stmt, 1));
  return user;
}

blist_user *blist_user_get_by_id(const blist *ctx, u_int id) {
  sqlite3_stmt *stmt = ctx->stmts[GET_USER];
  blist_user *user = NULL;
  sqlite3_bind_int(stmt, 1, id);
  if (sqlite3_step(stmt) == SQLITE_ROW)
    return blist_user_init_from_sql(stmt);
  sqlite3_reset(stmt);
  return user;
}

blist_user *blist_user_get_by_local(const blist *ctx, u_int local_id) {
  sqlite3_stmt *stmt = ctx->stmts[GET_USER_BY_LOCAL];
  blist_user *user = NULL;
  sqlite3_bind_int(stmt, 1, local_id);
  if (sqlite3_step(stmt) == SQLITE_ROW) {
    user = blist_user_init_from_sql(stmt);
  } else
    blist_print(ctx->log_ctx, BLIST_LOG_ERROR,
                "failed to get user by local id %u: %s\n", local_id,
                sqlite3_errmsg(ctx->db));
  sqlite3_reset(stmt);

  return user;
}

void blist_user_save(const blist *ctx, blist_user *user) {
  sqlite3_stmt *stmt = ctx->stmts[ADD_USER];
  int result = 0;
  if (user->id > -1)
    sqlite3_bind_int(stmt, 1, user->id);
  sqlite3_bind_text(stmt, 2, (char *)user->name, -1, SQLITE_STATIC);

  if ((result = sqlite3_step(stmt)) == SQLITE_ROW)
    user->id = sqlite3_column_int(stmt, 0);
  else
    blist_print(ctx->log_ctx, BLIST_LOG_ERROR, "failed to save user: %s - %d\n",
                sqlite3_errmsg(ctx->db), result);
  sqlite3_reset(stmt);

  if (user->local_id != -1) {
    stmt = ctx->stmts[ADD_LOGIN];
    sqlite3_bind_int(stmt, 1, user->id);
    sqlite3_bind_int(stmt, 2, user->local_id);
    if (sqlite3_step(stmt) != SQLITE_ROW)
      blist_print(ctx->log_ctx, BLIST_LOG_ERROR,
                  "failed to save logins for user: %s - %d\n",
                  sqlite3_errmsg(ctx->db), result);
    sqlite3_reset(stmt);
  } else
    blist_print(ctx->log_ctx, BLIST_LOG_ERROR, "no local uid saved for user %s",
                (char *)user->name);
}

void blist_user_delete(const blist *ctx, const blist_user *user) {
  sqlite3_stmt *stmt = ctx->stmts[DEL_USER];
  sqlite3_bind_int(stmt, 1, (int)user->id);
  if (sqlite3_step(stmt) == SQLITE_ROW)
    sqlite3_reset(stmt);
}

void blist_user_deinit(blist_user *user) {
  if (user->name)
    free(user->name);
  free(user);
}

blist_stage *blist_stage_init() {
  blist_stage *stage = malloc(sizeof(blist_stage));
  stage->id = -1;
  stage->name = NULL;
  stage->desc = NULL;
  stage->list_id = -1;
  return stage;
}

blist_stage *blist_stage_init_from_sql(sqlite3_stmt *stmt) {
  blist_stage *stage = blist_stage_init();
  stage->id = sqlite3_column_int(stmt, 0);
  if (sqlite3_column_bytes(stmt, 1) > 0)
    stage->name = (unsigned char *)strdup((char *)sqlite3_column_text(stmt, 1));
  if (sqlite3_column_bytes(stmt, 2) > 0)
    stage->desc = (unsigned char *)strdup((char *)sqlite3_column_text(stmt, 2));
  return stage;
}

blist_stage *blist_stage_get_by_id(const blist *ctx, u_int id) {
  sqlite3_stmt *stmt = ctx->stmts[GET_STAGE];
  sqlite3_bind_int(stmt, 1, id);
  if (sqlite3_step(stmt) == SQLITE_ROW) {
    sqlite3_reset(stmt);
    return blist_stage_init_from_sql(stmt);
  }
  return NULL;
}

blist_stage **blist_stage_get_all_for_list(const blist *ctx, u_int id,
                                           size_t *count) {
  sqlite3_stmt *stmt = ctx->stmts[GET_STAGES_FOR_LIST];
  sqlite3_bind_int(stmt, 1, id);
  blist_stage **stages = NULL;
  size_t stages_len = 0;
  while (sqlite3_step(stmt) == SQLITE_ROW) {
    blist_stage **_stages =
        realloc(stages, ++stages_len * sizeof(blist_stage *));
    if (_stages) {
      stages = _stages;
      stages[stages_len - 1] = blist_stage_init_from_sql(stmt);
    }
  }
  if (count != NULL)
    *count = stages_len;
  return stages;
}

void blist_stage_save(const blist *ctx, blist_stage *stage) {
  sqlite3_stmt *stmt = ctx->stmts[ADD_STAGE];
  int result = 0;
  if (stage->id > -1)
    sqlite3_bind_int(stmt, 1, stage->id);
  sqlite3_bind_text(stmt, 2, (char *)stage->name, -1, SQLITE_STATIC);
  sqlite3_bind_text(stmt, 3, (char *)stage->name, -1, SQLITE_STATIC);
  sqlite3_bind_int(stmt, 4, stage->list_id);
  if ((result = sqlite3_step(stmt)) == SQLITE_ROW || result == SQLITE_DONE) {
    stage->id = sqlite3_column_int(stmt, 0);
    sqlite3_reset(stmt);
  } else
    blist_print(ctx->log_ctx, BLIST_LOG_ERROR,
                "failed to save stage: %s - %d\n", sqlite3_errmsg(ctx->db),
                result);
}

void blist_stage_delete(const blist *ctx, const blist_stage *stage) {
  sqlite3_stmt *stmt = ctx->stmts[DEL_STAGE];
  sqlite3_bind_int(stmt, 1, stage->id);
  if (sqlite3_step(stmt) == SQLITE_DONE)
    sqlite3_reset(stmt);
  else
    blist_print(ctx->log_ctx, BLIST_LOG_ERROR, "failed to delete task: %s\n",
                sqlite3_errmsg(ctx->db));
}

void blist_stage_deinit(blist_stage *stage) {
  if (stage->name)
    free(stage->name);
  if (stage->desc)
    free(stage->desc);
  free(stage);
}

blist_list *blist_list_init() {
  blist_list *list = malloc(sizeof(blist_list));
  list->id = -1;
  list->name = NULL;
  list->desc = NULL;
  list->is_preset = false;
  return list;
}

blist_list *blist_list_init_from_sql(sqlite3_stmt *stmt) {
  blist_list *list = blist_list_init();
  list->id = sqlite3_column_int(stmt, 0);
  if (sqlite3_column_bytes(stmt, 1) > 0)
    list->name = (unsigned char *)strdup((char *)sqlite3_column_text(stmt, 1));
  if (sqlite3_column_bytes(stmt, 2) > 0)
    list->desc = (unsigned char *)strdup((char *)sqlite3_column_text(stmt, 2));

  list->is_preset = sqlite3_column_int(stmt, 3) ? true : false;
  return list;
}

blist_list *blist_list_get_by_id(const blist *ctx, u_int id) {
  sqlite3_stmt *stmt = ctx->stmts[GET_LIST];
  sqlite3_bind_int(stmt, 1, id);
  if (sqlite3_step(stmt) == SQLITE_ROW) {
    sqlite3_reset(stmt);
    return blist_list_init_from_sql(stmt);
  }
  return NULL;
}

blist_list **blist_list_get_all_by_user(const blist *ctx, u_int user_id,
                                        size_t *count) {
  sqlite3_stmt *stmt = ctx->stmts[GET_LISTS_BY_USER];
  sqlite3_bind_int(stmt, 1, user_id);
  blist_list **lists = NULL;
  size_t lists_len = 0;
  while (sqlite3_step(stmt) == SQLITE_ROW) {
    blist_list **_lists = realloc(lists, ++lists_len * sizeof(blist_list *));
    if (_lists) {
      lists = _lists;
      lists[lists_len - 1] = blist_list_init_from_sql(stmt);
    }
  }
  if (count != NULL)
    *count = lists_len;
  return lists;
}

blist_list *blist_list_get_by_user_and_name(const blist *ctx, u_int user_id,
                                            const char *name) {
  sqlite3_stmt *stmt = ctx->stmts[GET_LISTS_BY_USER_NAME];
  blist_list *list = NULL;
  sqlite3_bind_int(stmt, 1, user_id);
  sqlite3_bind_text(stmt, 2, name, -1, SQLITE_STATIC);

  if (sqlite3_step(stmt) == SQLITE_ROW)
    list = blist_list_init_from_sql(stmt);
  else
    blist_print(ctx->log_ctx, BLIST_LOG_ERROR,
                "failed to get lists for user: %s\n", sqlite3_errmsg(ctx->db));
  sqlite3_reset(stmt);
  return list;
}

void blist_list_save(const blist *ctx, blist_list *list,
                     const blist_user *user) {
  sqlite3_stmt *stmt = ctx->stmts[ADD_LIST];
  int result = 0;
  if (list->id > -1)
    sqlite3_bind_int(stmt, 1, list->id);
  sqlite3_bind_text(stmt, 2, (char *)list->name, -1, SQLITE_STATIC);
  sqlite3_bind_text(stmt, 3, (char *)list->desc, -1, SQLITE_STATIC);
  sqlite3_bind_int(stmt, 4, list->is_preset);
  if ((result = sqlite3_step(stmt)) == SQLITE_ROW)
    list->id = sqlite3_column_int(stmt, 0);
  else
    blist_print(ctx->log_ctx, BLIST_LOG_ERROR, "failed to save list: %s - %d\n",
                sqlite3_errmsg(ctx->db), result);
  sqlite3_reset(stmt);

  if (list->id > -1) {
    stmt = ctx->stmts[ADD_USER_LIST];
    sqlite3_bind_int(stmt, 1, user->id);
    sqlite3_bind_int(stmt, 2, list->id);
    if (sqlite3_step(stmt) == SQLITE_DONE)
      sqlite3_reset(stmt);
    else
      blist_print(ctx->log_ctx, BLIST_LOG_ERROR,
                  "failed to save user - list relation in join table: %s\n",
                  sqlite3_errmsg(ctx->db));
  }
}

void blist_list_delete(const blist *ctx, blist_list *list) {
  sqlite3_stmt *stmt = ctx->stmts[DEL_LIST];
  sqlite3_bind_int(stmt, 1, list->id);
  if (sqlite3_step(stmt) == SQLITE_DONE)
    sqlite3_reset(stmt);
  else
    blist_print(ctx->log_ctx, BLIST_LOG_ERROR, "failed to delete task: %s\n",
                sqlite3_errmsg(ctx->db));
}

void blist_list_deinit(blist_list *list) {
  if (list->name)
    free(list->name);
  if (list->desc)
    free(list->desc);
  free(list);
}

blist_task *blist_task_init() {
  blist_task *task = malloc(sizeof(blist_task));
  task->id = -1;
  task->name = NULL;
  task->desc = NULL;
  task->due = -1;
  task->target_stage = -1;
  return task;
}

blist_task *blist_task_init_from_sql(sqlite3_stmt *stmt) {
  blist_task *task = blist_task_init();
  task->id = sqlite3_column_int(stmt, 0);
  if (sqlite3_column_bytes(stmt, 1) > 0)
    task->name = (unsigned char *)strdup((char *)sqlite3_column_text(stmt, 1));
  if (sqlite3_column_bytes(stmt, 2) > 0)
    task->desc = (unsigned char *)strdup((char *)sqlite3_column_text(stmt, 2));
  task->stage_id = sqlite3_column_int(stmt, 3);
  task->due = sqlite3_column_int(stmt, 4);
  task->target_stage = sqlite3_column_int(stmt, 5);
  return task;
}

blist_task *blist_task_get_by_id(const blist *ctx, u_int id) {
  sqlite3_stmt *stmt = ctx->stmts[GET_TASK];
  sqlite3_bind_int(stmt, 1, id);
  if (sqlite3_step(stmt) == SQLITE_ROW) {
    sqlite3_reset(stmt);
    return blist_task_init_from_sql(stmt);
  }
  return NULL;
}

blist_task **blist_task_get_for_list(const blist *ctx, u_int list_id,
                                     size_t *count) {
  sqlite3_stmt *stmt = ctx->stmts[GET_TASKS_FOR_LIST];
  sqlite3_bind_int(stmt, 1, list_id);
  blist_task **tasks = NULL;
  size_t tasks_len = 0;
  while (sqlite3_step(stmt) == SQLITE_ROW) {
    tasks_len++;
    blist_task **_tasks = realloc(tasks, tasks_len * sizeof(blist_task *));
    if (_tasks) {
      tasks = _tasks;
      tasks[tasks_len - 1] = blist_task_init_from_sql(stmt);
    }
  }
  sqlite3_reset(stmt);
  *count = tasks_len;
  return tasks;
}

blist_task *blist_task_get_for_list_by_name(const blist *ctx, u_int list_id,
                                            const char *name) {
  sqlite3_stmt *stmt = ctx->stmts[GET_TASKS_FOR_LIST_BY_NAME];
  blist_task *task = NULL;
  sqlite3_bind_int(stmt, 1, list_id);
  sqlite3_bind_text(stmt, 2, name, -1, SQLITE_STATIC);
  if (sqlite3_step(stmt) == SQLITE_ROW)
    task = blist_task_init_from_sql(stmt);
  else
    blist_print(ctx->log_ctx, BLIST_LOG_WARNING,
                "no tasks found for list %d, name '%s': %s\n", list_id, name,
                sqlite3_errmsg(ctx->db));
  sqlite3_reset(stmt);
  return task;
}

void blist_task_save(const blist *ctx, blist_task *task) {
  sqlite3_stmt *stmt = ctx->stmts[ADD_TASK];
  int result = 0;
  if (task->id > -1)
    sqlite3_bind_int(stmt, 1, task->id);
  sqlite3_bind_text(stmt, 2, (char *)task->name, -1, SQLITE_STATIC);
  sqlite3_bind_text(stmt, 3, (char *)task->desc, -1, SQLITE_STATIC);
  sqlite3_bind_int(stmt, 4, task->stage_id);
  if (task->due > 0)
    sqlite3_bind_int(stmt, 5, task->due);
  if (task->target_stage > 0)
    sqlite3_bind_int(stmt, 6, task->target_stage);
  if ((result = sqlite3_step(stmt)) == SQLITE_ROW) {
    task->id = sqlite3_column_int(stmt, 0);
  } else
    blist_print(ctx->log_ctx, BLIST_LOG_ERROR, "failed to save list: %s - %d\n",
                sqlite3_errmsg(ctx->db), result);
  sqlite3_reset(stmt);
}
void blist_task_delete(const blist *ctx, const blist_task *task) {
  sqlite3_stmt *stmt = ctx->stmts[DEL_TASK];
  sqlite3_bind_int(stmt, 1, task->id);
  if (sqlite3_step(stmt) == SQLITE_DONE)
    sqlite3_reset(stmt);
  else
    blist_print(ctx->log_ctx, BLIST_LOG_ERROR, "failed to delete task: %s\n",
                sqlite3_errmsg(ctx->db));
}

void blist_task_deinit(blist_task *task) {
  if (task->name)
    free(task->name);
  if (task->desc)
    free(task->desc);
  free(task);
}

/*
 * printing
 */

void blist_print_user(blist *ctx, const blist_user *user) {
  fprintf(ctx->log_ctx->log_target,
          ""
          "Name: %s\n"
          "has local account: %s\n",
          (char *)user->name, user->local_id > 0 ? "yes" : "no");
}

void print_table_line(char *left, size_t cols, char *filler, size_t col_width,
                      char *col_delim, char *right) {
  printf("%s", left);
  int i;
  for (i = 0; i < cols; i++) {
    int j = 0;
    for (; j < col_width; j++)
      printf("%s", filler);
    if (i + 1 != cols)
      printf("%s%s", filler, col_delim);
  }
  printf("%s\n", right);
}

void blist_print_list(const blist *ctx, const blist_list *list,
                      blist_task **tasks, size_t tasks_len) {
  struct winsize w;
  int i = 0;
  ioctl(STDOUT_FILENO, TIOCGWINSZ, &w);
  u_int width = w.ws_col > 0 ? w.ws_col : 40;
  for (i = 0; i < tasks_len; i++)
    fprintf(stderr, "%d, %s\n", i, (char *)tasks[i]->name);

  size_t stages_len = 0;
  blist_stage **stages =
      blist_stage_get_all_for_list(ctx, list->id, &stages_len);

  int max_stages_id = 0;
  for (i = 0; i < stages_len; i++)
    if (stages[i]->id > max_stages_id)
      max_stages_id = stages[i]->id;

  int tasks_count_per_stage[max_stages_id + 1];
  for (i = 0; i < max_stages_id + 1; i++)
    tasks_count_per_stage[i] = 0;

  u_long task_length = 0, max_task_length = 0;
  for (i = 0; i < tasks_len; i++) {
    tasks_count_per_stage[tasks[i]->stage_id]++;
    task_length = strlen((char *)tasks[i]->name);
    if (task_length > max_task_length)
      max_task_length = task_length;
  }

  u_int content_width =
      (int)((double)max_task_length * 1.6 * (double)stages_len);
  if (content_width)
    width = content_width;
  size_t stage_col_width = (width - 3 - (stages_len - 1)) / stages_len;

  int tasks_max = 0;
  for (i = 0; i < max_stages_id + 1; i++)
    if (tasks_count_per_stage[i] > tasks_max)
      tasks_max = tasks_count_per_stage[i];

  print_table_line("╭", stages_len, "─", stage_col_width, "─", "╮");

  // list name
  printf("│");
  printf(" %s", (char *)list->name);
  for (i = 0;
       i < (stages_len * stage_col_width) - (strlen((char *)list->name + 1));
       i++)
    printf(" ");
  printf("│\n");

  // separator
  print_table_line("├", stages_len, "─", stage_col_width, "┬", "┤");

  // stage name
  printf("│");
  for (i = 0; i < stages_len; i++) {
    int j = 0;
    printf(" %s", (char *)stages[i]->name);
    for (; j < (stage_col_width - 1) - strlen((char *)stages[i]->name); j++)
      printf(" ");
    if (i + 1 != stages_len)
      printf(" │");
  }
  printf("│\n");

  /*
  const char *crossings[] = { "┬", "┼", "┴" };
  const char *headers[] = { "Name", "due" };
  for (i = 0; i < 3; i++) {
    printf("%s", (i + 1) % 2 ? "├" : "│");
    int j = 0;
    for (j = 0; j < stages_len * 2; j++) {
      int k = 0, printed = 0;
      if (i == 1) {
        printf(" %s", headers[j % 2]);
        printed = strlen(headers[j % 2]) + 1;
      }
      for (; k < ((stage_col_width / 2) - 1) - printed; k++)
        printf("%s", (i + 1) % 2 ? "─" : " ");
      if (j + 1 != stages_len * 2)
        if ((i + 1) % 2)
          printf("─%s", j % 2 ? crossings[1] : crossings[i > 1]);
        else
          printf(" │");
    }
    printf("%s\n", (i + 1) % 2 ? "┤" : "│");
  }
  */
  // separator
  if (tasks_len)
    print_table_line("├", stages_len, "─", stage_col_width, "┼", "┤");

  // tasks
  int stages_idx[stages_len];
  for (i = 0; i < stages_len; i++)
    stages_idx[i] = 0;
  for (i = 0; i < tasks_max; i++) {
    printf("│");
    int j = 0;
    for (; j < stages_len; j++) {
      int k = stages_idx[j];
      u_long printed = 0;
      for (; k < tasks_len; k++)
        if (tasks[k]->stage_id == stages[j]->id) {
          printed = printf(" %s", (char *)tasks[k]->name);
          stages_idx[j] = k + 1;
          break;
        }
      int l = 0;
      for (; l < stage_col_width - printed; l++)
        printf(" ");
      if (j + 1 != stages_len)
        printf(" │");
    }
    printf("│\n");
  }

  // end line
  print_table_line("╰", stages_len, "─", stage_col_width, "┴", "╯");

  // cleanup
  for (i = 0; i < stages_len; i++)
    blist_stage_deinit(stages[i]);
  free(stages);
}
