#include <stdio.h>
#include <stdlib.h>

#include "util.h"

#include <string.h>

#include "models.h"
#include "sql.h"

/*
 * command parsing
 */

blist *blist_init(char *db) {
  blist *ctx = malloc(sizeof(blist));
  ctx->stmts = malloc(_BLIST_SQL_COUNT * sizeof(sqlite3_stmt *));
  ctx->error = 0;
  ctx->log_ctx = malloc(sizeof(blist_logging_ctx));
  blist_sql_prepare(ctx, db);

  return ctx;
}
void blist_deinit(blist *ctx) {
  int i = 0;
  for (; i < _BLIST_SQL_COUNT; i++)
    sqlite3_finalize(ctx->stmts[i]);
  sqlite3_close(ctx->db);
  if (ctx->log_ctx) {
    if (ctx->log_ctx->log_target)
      fclose(ctx->log_ctx->log_target);
    free(ctx->log_ctx);
  }
  free(ctx->stmts);
  free(ctx);
  ctx = NULL;
}

/*
 * sql preparations
 */

void blist_sql_prepare(blist *ctx, char *db) {
  if (!db)
    blist_print(
        ctx->log_ctx, BLIST_LOG_WARNING,
        "Database will be in-memory only, changes are not persistent!\n");
  if (sqlite3_open(db ? db : ":memory:", &ctx->db) != SQLITE_OK) {
    blist_print(ctx->log_ctx, BLIST_LOG_ERROR, "Can't open database: %s\n",
                sqlite3_errmsg(ctx->db));
    return;
  }
  char *errmsg = NULL;
  sqlite3_exec(ctx->db, blist_sql_get(INIT), NULL, NULL, &errmsg);
  if (errmsg) {
    blist_print(ctx->log_ctx, BLIST_LOG_ERROR,
                "Database initialization failed (%s), expect issues.\n",
                sqlite3_errmsg(ctx->db));
    sqlite3_free(errmsg);
  }
  int i = 0;
  for (; i < _BLIST_SQL_COUNT; i++)
    if (sqlite3_prepare(ctx->db, blist_sql_get(i), -1, &ctx->stmts[i], NULL) !=
        SQLITE_OK)
      blist_print(ctx->log_ctx, BLIST_LOG_WARNING,
                  "sqlite3_prepare for '%s' failed: %s\n", blist_sql_get(i),
                  sqlite3_errmsg(ctx->db));
}

void blist_sql_seed(blist *ctx) {}

/*
 * assuring things
 */

bool blist_assure_user(const blist *ctx, const __uid_t id, const char *name) {
  blist_user *user = blist_user_get_by_local(ctx, id);
  bool created = false;
  if (!user) {
    user = blist_user_init();
    user->name = malloc((strlen(name) + 1) * sizeof(char));
    strcpy((char *)user->name, name);
    user->local_id = id;
    blist_user_save(ctx, user);
    created = true;
  }
  blist_user_deinit(user);
  return created;
}

/*
 * printing methods
 */

void blist_print(blist_logging_ctx *ctx, BLIST_LOG_LEVEL log_level,
                 const char *format, ...) {
  if (ctx->log_level >= log_level) {
    switch (log_level) {
    case BLIST_LOG_DEBUG:
      fprintf(ctx->log_target, "[\x1B[34mDBG\x1B[0m] ");
      break;
    case BLIST_LOG_INFO:
      fprintf(ctx->log_target, "[\x1B[36mINF\x1B[0m] ");
      break;
    case BLIST_LOG_WARNING:
      fprintf(ctx->log_target, "[\x1B[33mWRN\x1B[0m] ");
      break;
    case BLIST_LOG_ERROR:
      fprintf(ctx->log_target, "[\x1B[31mERR\x1B[0m] ");
      break;
    }
    va_list args;
    va_start(args, format);
    vprintf(format, args);
    va_end(args);
  }
}